#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include "types.hpp"
#include "types_isd.hpp"
#include "fmtReader.hpp"
#include "tools.hpp"

FmtReader::FmtReader(const wchar_t *path)
{
  noPlugins = 0;
  wchar_t pluginPath[300];
  wchar_t pluginMask[300];
  
  lstrcpyW(pluginPath, path);
  lstrcpyW(pointToName(pluginPath), L"Formats\\");
  
  lstrcpyW(pluginMask, pluginPath);
  lstrcatW(pluginMask, L"*.fmt");

  WIN32_FIND_DATAW data;
  HANDLE foundFile = FindFirstFileW(pluginMask, &data);
  
  int done = (foundFile == INVALID_HANDLE_VALUE);
  while(!done)
  {
    wchar_t pluginName[300];
    lstrcpyW(pluginName, pluginPath);
    lstrcatW(pluginName, data.cFileName);

    HINSTANCE hPlugin = LoadLibraryW(pluginName);
    if(hPlugin != NULL)
    {
      plugins[noPlugins].handle         = hPlugin;
      plugins[noPlugins].isImage        = (IS_IMAGE)         GetProcAddress(hPlugin,"isImage");
      plugins[noPlugins].openSubPlugin  = (OPEN_SUB_PLUGIN)  GetProcAddress(hPlugin,"openSubPlugin");
      plugins[noPlugins].closeSubPlugin = (CLOSE_SUB_PLUGIN) GetProcAddress(hPlugin,"closeSubPlugin");
      plugins[noPlugins].reload         = (RELOAD)           GetProcAddress(hPlugin,"reload");
      plugins[noPlugins].openFile       = (OPEN_FILE)        GetProcAddress(hPlugin,"openFile");
      plugins[noPlugins].closeFile      = (CLOSE_FILE)       GetProcAddress(hPlugin,"closeFile");
      plugins[noPlugins].read           = (READ)             GetProcAddress(hPlugin,"read");
      plugins[noPlugins].write          = (WRITE)            GetProcAddress(hPlugin,"write");
      plugins[noPlugins].getFormatName  = (GET_FORMAT_NAME)  GetProcAddress(hPlugin,"getFormatName");

      if(plugins[noPlugins].isImage        != NULL &&
         plugins[noPlugins].openSubPlugin  != NULL &&
         plugins[noPlugins].closeSubPlugin != NULL &&
         plugins[noPlugins].reload         != NULL &&
         plugins[noPlugins].openFile       != NULL &&
         plugins[noPlugins].closeFile      != NULL &&
         plugins[noPlugins].read           != NULL &&
         plugins[noPlugins].write          != NULL &&
         plugins[noPlugins].getFormatName  != NULL) ++noPlugins;
      
      if(noPlugins == 16) break;
    }
    done = !FindNextFileW(foundFile, &data);
  }
  FindClose(foundFile);
}

FmtPlugin* FmtReader::isImage(const wchar_t *fileName, const unsigned char *data, int size)
{
  for(int i = 0; i < noPlugins; ++i)
    if(plugins[i].isImage(fileName, data, size)) return &plugins[i];
  return NULL;
}
