#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <string>

#include "types.hpp"
#include "types_isd.hpp"
#include "manager.hpp"
#include "tools.hpp"
#include "tools_isd.hpp"
#include "iSDOS.hpp"
#include "iSDOS_tools.hpp"
#include "iterator.hpp"
#include "lang.hpp"

extern PluginStartupInfo startupInfo;
extern const GUID MainGuid;

int Manager::makeFolder(UniHdr* pDir, const wchar_t* name)
{
  bool isChic;
  if(dsk.signatureChic[0] == 'D' && dsk.signatureChic[1] == 'S' && dsk.signatureChic[2] == 'K')
    isChic = true;
  else
    isChic = false;
  if(isChic)
  {
    if(pDir[0].dir.level == 6 ||
       pDir[0].dir.levelChic == 6 ||
       pDir[0].dir.noFiles == 127) return 0;
  }
  else
    if(pDir[0].dir.level == 6 ||
       pDir[0].dir.noFiles == 127) return 0;
  if(!reserveSpaceForFiles(pDir, 1)) return -1;
  if(noFreeBlocks < 2) return -1;

  UniHdr newDir{};
  std::fill(std::begin(newDir.name), std::end(newDir.name), ' ');
  std::fill(std::begin(newDir.ext), std::end(newDir.ext), ' ');

  for(int i = 0; i < 8; ++i)
  {
    if(!name[i]) break;

    char c;
    //   cp866
    WideCharToMultiByte(866, 0, &name[i], 1, &c, 1, nullptr, nullptr);

    newDir.name[i] = isValidCharIsdos(c) ? c : '_';
  }
  
  newDir.attr                  = 0x21;
  newDir.dir.parentDir1stBlock = pDir[0].dir.firstBlock;
  newDir.dir.size              = blockSize;
  if(isChic)
    if(pDir[0].dir.level != 0)
      newDir.dir.level         = pDir[0].dir.level + 1;
    if(pDir[0].dir.levelChic != 0)
      newDir.dir.levelChic     = pDir[0].dir.levelChic + 1;
    if(pDir[0].dir.level == 0 && pDir[0].dir.levelChic == 0)
      newDir.dir.levelChic     = pDir[0].dir.levelChic + 1;
  else
    newDir.dir.level           = pDir[0].dir.level + 1;

  newDir.dir.firstBlock        = findFreeBlocks(1);
  markBlock(newDir.dir.firstBlock);

  newDir.dir.descr1stBlock     = findFreeBlocks(1);
  markBlock(newDir.dir.descr1stBlock);

  u8 block[blockSize]{};
  block[0] = 1;
  *(u16*)(block+1) = newDir.dir.firstBlock;
  block[3] = 1;
  if(!writeBlock(newDir.dir.descr1stBlock, block))
  {
    read_device_sys();
    return -1;
  }
  
  newDir.dir.totalFiles        = 1;
  newDir.dir.noFiles           = 0;

  SYSTEMTIME time;
  GetLocalTime(&time);
  setDate(newDir, time);

  u8 buf[blockSize]{};
  CopyMemory(buf, &newDir, sizeof(UniHdr));
  writeFolder((UniHdr*)buf);
  int fNum = 1;
  for(; fNum < pDir[0].dir.totalFiles; ++fNum)
    if(!(pDir[fNum].attr & FLAG_EXIST)) break;
  
  CopyMemory(&pDir[fNum], &newDir, sizeof(UniHdr));
  if(fNum == pDir[0].dir.totalFiles) ++pDir[0].dir.totalFiles;
  ++pDir[0].dir.noFiles;

  writeFolder(pDir);
  write_device_sys();
  return 1;
}

int Manager::makeDirectory(const wchar_t *dirName, int opMode)
{
  FarDialogItem dialogItems[] =
  {
    { DI_DOUBLEBOX, 3, 1, 60, 6, { FALSE },             nullptr, nullptr,                                   0, GetMsg(MMakeFolder) },
    { DI_TEXT,      5, 2,  0, 0, { FALSE },             nullptr, nullptr,                                   0, GetMsg(MCreateFolder) },
    { DI_EDIT,      5, 3, 58, 3,  { TRUE }, L"XiSD_folder_name", nullptr,                         DIF_HISTORY, dirName, 8},
    { DI_TEXT,      3, 4,  0, 0, { FALSE },             nullptr, nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L""},
    { DI_BUTTON,    0, 5,  0, 0, { FALSE },             nullptr, nullptr, DIF_CENTERGROUP | DIF_DEFAULTBUTTON, GetMsg(MOk) },
    { DI_BUTTON,    0, 5,  0, 0, { FALSE },             nullptr, nullptr,                     DIF_CENTERGROUP, GetMsg(MCancel) }
  };

  //    ,  
  std::wstring DirNameStorage;
  if((opMode & OPM_SILENT) == 0)
  {
    std::unique_ptr<void, decltype(SafeDlgFree) *> Dlg{ startupInfo.DialogInit(&MainGuid,
                                   &__uuidof(struct __declspec(uuid("{D40D86D0-63E9-4058-BA2C-D67F131411B0}")) TDlgGuid),
                                   -1, -1, 64, 8,
                                   nullptr,
                                   dialogItems,
                                   std::size(dialogItems), 0, 0, startupInfo.DefDlgProc, nullptr), SafeDlgFree };
    if(Dlg.get() == INVALID_HANDLE_VALUE)
    {
        return FALSE;
    }
    intptr_t askCode = startupInfo.DialogRun(Dlg.get());

    if(askCode != 4) return -1;

    auto Sz = startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 2, nullptr);
    DirNameStorage.resize(Sz);

    FarDialogItemData ItemData = { sizeof(FarDialogItemData), DirNameStorage.size(), DirNameStorage.data() };
    startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 2, &ItemData);
    dirName = DirNameStorage.c_str();
  }

  if(*dirName == 0) return 1;

  wchar_t cname[8+3+2] = L"";
  makeCompatibleFolderName(cname, dirName);

  int fNum = findFile(files, cname);
  if(fNum) return 0;
  if(!openHostFile()) return 0;
  int result = makeFolder(files, cname);

  closeHostFile();
  startupInfo.PanelControl(PANEL_PASSIVE, FCTL_UPDATEPANEL, 1, nullptr);
  return result;
}
