#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include "types.hpp"
#include "types_isd.hpp"
#include "manager.hpp"
#include "tools.hpp"
#include "tools_isd.hpp"
#include "iSDOS.hpp"
#include "iSDOS_tools.hpp"
#include "iterator.hpp"
#include "lang.hpp"

extern PluginStartupInfo startupInfo;
extern Options           op;

int Manager::putOneFolder(UniHdr* pDir, const wchar_t *FileName, const wchar_t* fromDir, const wchar_t* dirName, int move, Action& action, int opMode)
{
  wchar_t cname[8+3+2] = L"";
  makeCompatibleFolderName(cname, FileName);
  
  int fNum  = findFile(pDir, cname);
  if(!fNum)
  {
    int result = makeFolder(pDir, cname);
    if(result != 1) return result;
  }
  UniHdr curDir[128]{};
  fNum  = findFile(pDir, cname);
  readFolder(pDir[fNum], curDir);

  wchar_t fullName[300];
  makeFullName(fullName, dirName, FileName);

  wchar_t fullFromName[300];
  makeFullName(fullFromName, fromDir, FileName);

  wchar_t mask[300];
  makeFullName(mask, fullFromName, L"*.*");
  
  WIN32_FIND_DATAW data;
  HANDLE h = FindFirstFileW(mask, &data);
  if(h == INVALID_HANDLE_VALUE) return 1;
  int exitCode = 1;
  do
  {
    int result;
    if(!lstrcmpW(data.cFileName, L".")) continue;
    if(!lstrcmpW(data.cFileName, L"..")) continue;
    if(!reserveSpaceForFiles(curDir, 1))
    {
      exitCode = -1;
      break;
    }
    if(data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
      result = putOneFolder(curDir, data.cFileName, fullFromName, fullName, move, action, opMode);
    else
      result = putOneFile(curDir, data, fullFromName, fullName, move, action, opMode);

    if(result != 1)
    {
      exitCode = result;
      if(exitCode == -1) break;
    }
   
  } while(FindNextFileW(h, &data));
  
  FindClose(h);
  writeFolder(curDir);
  if(move && exitCode == 1)
    if(!RemoveDirectoryW(fullFromName)) return 0;
  return exitCode;
}

int Manager::putOneFile(UniHdr* pDir, const WIN32_FIND_DATAW &file, const wchar_t* fromDir, const wchar_t* dirName, int move, Action& action, int opMode)
{

  u16 needBlocks = file.nFileSizeLow/blockSize + (file.nFileSizeLow%blockSize ? 1 : 0);
  int firstFoundBlock = findFreeBlocks(needBlocks);
  
  if(firstFoundBlock == -1)
  {
    const wchar_t *msgItems[3];
    msgItems[0] = GetMsg(MError);
    msgItems[1] = GetMsg(MNoDiskSpace);
    msgItems[2] = GetMsg(MOk);
    messageBox(FMSG_WARNING, msgItems, sizeof(msgItems)/sizeof(msgItems[0]), 1);
    return -1;
  }
  wchar_t cname[8+3+2] = L"";
  makeCompatibleFileName(cname, file.cFileName);

  int fNum = findFile(pDir, cname);
  if(fNum)
  {
    //        
    if(isDir(pDir[fNum]))  return 0; //  
    if(pDir[fNum].attr == 0xFF && pDir[fNum].file.systemFlag == 0xFF) return 0;
    
    if(action == SKIP_ALL) return 0;
    if(action == ASK_USER)
    {
      wchar_t fullName[300];
      makeFullName(fullName, dirName, file.cFileName);

      const wchar_t *msgItems[12];
      msgItems[0] = GetMsg(MWarning);
      msgItems[1] = GetMsg(MAlreadyExists);
      msgItems[3] = L"\x001";
      msgItems[6] = L"\x001";
      msgItems[7] = GetMsg(MOverwrite);
      msgItems[8] = GetMsg(MAll);
      msgItems[9] = GetMsg(MSkip);
      msgItems[10] = GetMsg(MSkipAll);
      msgItems[11] = GetMsg(MCancel);
      
      wchar_t shortName[50];
      msgItems[2] = cropLongName(shortName, fullName, 47);

      wchar_t dest  [50];
      wchar_t source[50];

      FILETIME lastWriteTime;
      FileTimeToLocalFileTime(&file.ftLastWriteTime, &lastWriteTime);
  
      SYSTEMTIME time;
      FileTimeToSystemTime(&lastWriteTime, &time);
      
      wsprintfW(source, GetMsg(MSrc),
               file.nFileSizeLow,
               time.wDay, time.wMonth, time.wYear,
               time.wHour, time.wMinute, time.wSecond);
      
      SYSTEMTIME st = makeDate(pDir[fNum]);
      wsprintfW(dest, GetMsg(MDest),
               getSize(pDir[fNum]),
               st.wDay, st.wMonth, st.wYear,
               st.wHour, st.wMinute, st.wSecond);
      
      msgItems[4] = source;
      msgItems[5] = dest;
      int askCode = messageBox(FMSG_WARNING /*| FMSG_DOWN*/,
                               msgItems,
                               sizeof(msgItems)/sizeof(msgItems[0]), 5);
      switch(askCode)
      {
        case -1:
        case 4:
                  return -1;
        case 3:
                  action = SKIP_ALL;
        case 2:
                  return 0;
        case 1:
                  action = DO_ALL;
      }
    }
    //  
    Action actProtected = DO_ALL;
    Action actFolder    = DO_ALL;
    deleteOneFile(pDir[fNum], pDir, dirName, OPM_SILENT, actProtected, actFolder);
  }

  wchar_t fullName[300];
  makeFullName(fullName, fromDir, file.cFileName);
  if(!(opMode & OPM_SILENT))
  {
    const wchar_t *msgItems[3];
    if(move)
    {
      msgItems[0] = GetMsg(MMove);
      msgItems[1] = GetMsg(MMoving);
    }
    else
    {
      msgItems[0] = GetMsg(MCopy);
      msgItems[1] = GetMsg(MCopying);
    }
    
    wchar_t shortName[45];
    msgItems[2] = cropLongName(shortName, fullName, 40);
    messageBox(0, msgItems, sizeof(msgItems)/sizeof(msgItems[0]), 0);
  }
  
  HANDLE hFile = CreateFileW(fullName,
                            GENERIC_READ,
                            FILE_SHARE_READ | FILE_SHARE_WRITE,
                            NULL,
                            OPEN_EXISTING,
                            FILE_FLAG_SEQUENTIAL_SCAN,
                            NULL);
  
  if(hFile == INVALID_HANDLE_VALUE)
  {
    const wchar_t *msgItems[4];
    msgItems[0] = GetMsg(MError);
    msgItems[1] = GetMsg(MCanNotOpen);
    msgItems[2] = file.cFileName;
    msgItems[3] = GetMsg(MOk);
    messageBox(FMSG_WARNING, msgItems, sizeof(msgItems)/sizeof(msgItems[0]), 1);
    return 0;
  }

  UniHdr newFile{};

  FillMemory(&newFile, 8+3, ' ');
  const wchar_t *p = cname;
  if(*p == 0) return 0;
  for(int i = 0; i < 8; ++i)
  {
    if(*p == 0 || *p =='.') break;
      newFile.name[i] = *p;
    ++p;
  }
  
  p = pointToExt(cname);
  
  if(p)
  {
    for(int i = 0; i < 3; ++i)
    {
      if(*p == 0) break;
      newFile.ext[i] = *p;
    ++p;
    }
  }
  
  newFile.attr            = 0x43;
  newFile.file.firstBlock = firstFoundBlock;
  setFileSize(newFile, file.nFileSizeLow);

  SYSTEMTIME time;
  FILETIME lastWriteTime;
  FileTimeToLocalFileTime(&file.ftLastWriteTime, &lastWriteTime);
  FileTimeToSystemTime(&lastWriteTime, &time);
  setDate(newFile, time);

  FILETIME ftCreationTime;
  GetFileTime(hFile, &ftCreationTime, NULL, NULL);
  SYSTEMTIME tmp;
  FileTimeToSystemTime(&ftCreationTime, &tmp);
  if(tmp.wYear == 2000 && tmp.wSecond == 0 && (tmp.wMonth == 1 || tmp.wMonth == 2))
  {
    u16 laddr = tmp.wMinute;
    laddr    += 60*tmp.wHour;
    laddr    += 24*60*(tmp.wDay-1);
    laddr    += 31*24*60*(tmp.wMonth-1);
    newFile.file.loadAddr = laddr;
  }
  u8 buf[blockSize];
  u16 checkSum = 0;
  while(needBlocks--)
  {
    u32 noBytesRead;
    ZeroMemory(buf, blockSize);
    ReadFile(hFile, buf, blockSize, &noBytesRead, NULL);
    writeBlock(firstFoundBlock, buf);
    markBlock(firstFoundBlock++);
    for(int i = 0; i < noBytesRead; ++i) checkSum += buf[i];
  }
  if(newFile.ext[0] == 'c' && newFile.ext[1] == 'o' && newFile.ext[2] == 'm')
    newFile.file.crc = 0 - checkSum;
 
  fNum = 1;
  for(; fNum < pDir[0].dir.totalFiles; ++fNum)
    if(!(pDir[fNum].attr & FLAG_EXIST)) break;
  
  CopyMemory(&pDir[fNum], &newFile, sizeof(UniHdr));
  if(fNum == pDir[0].dir.totalFiles) ++pDir[0].dir.totalFiles;
  ++pDir[0].dir.noFiles;

  CloseHandle(hFile);
  if(move)
    if(!DeleteFileW(fullName)) return 0;
  return 1;
}

int Manager::putFiles(PluginPanelItem *panelItem, int noItems, int move, int opMode)
{
  if(noItems == 0) return 1;
  if(files[0].dir.noFiles + noItems > 127)
  {
    const wchar_t *msgItems[3];
    msgItems[0] = GetMsg(MError);
    msgItems[1] = GetMsg(MMany);
    msgItems[2] = GetMsg(MOk);
    messageBox(FMSG_WARNING, msgItems, sizeof(msgItems)/sizeof(msgItems[0]), 1);
    return 0;
  }

  if(!openHostFile()) return 0;
  if(!reserveSpaceForFiles(files, noItems))
  {
    closeHostFile();
    const wchar_t *msgItems[3];
    msgItems[0] = GetMsg(MError);
    msgItems[1] = GetMsg(MNoDiskSpace);
    msgItems[2] = GetMsg(MOk);
    messageBox(FMSG_WARNING, msgItems, sizeof(msgItems)/sizeof(msgItems[0]), 1);
    return 0;
  }

  HANDLE screen = startupInfo.SaveScreen(0,0,-1,-1);
  
  Action action   = ASK_USER;
  int    exitCode = 1;
  for(int iNum = 0; iNum < noItems; ++iNum)
  {
    int result;
    if(panelItem[iNum].FileAttributes & FILE_ATTRIBUTE_DIRECTORY)
      result = putOneFolder(files, panelItem[iNum].FileName, L"", curDir, move, action, opMode);
    else
    {
        WIN32_FIND_DATAW FindData{};
        wcscpy(FindData.cFileName, panelItem[iNum].FileName);
        FindData.ftLastWriteTime = panelItem[iNum].LastWriteTime;
        FindData.nFileSizeLow = panelItem[iNum].FileSize;
        FindData.nFileSizeHigh = panelItem[iNum].FileSize >> 32U;
        result = putOneFile(files, FindData, L"", curDir, move, action, opMode);
    }

    if(result == -1)
    {
      exitCode = -1;
      break;
    }
    if(result ==  0)
    {
      exitCode = 0;
      continue;
    }
    panelItem[iNum].Flags ^= PPIF_SELECTED;
    
  }

  startupInfo.RestoreScreen(screen);

  writeFolder(files);
  write_device_sys();
  closeHostFile();
  op.reread = true;
  startupInfo.PanelControl(this, FCTL_UPDATEPANEL, 1, nullptr);
  op.reread = false;
  return exitCode;
}
