#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <iterator>

#include "plugin.hpp"

#include "types.hpp"
#include "types_scl.hpp"
#include "manager.hpp"
#include "tools.hpp"
#include "lang.hpp"

extern PluginStartupInfo startupInfo;
extern Options           op;

bool Manager::deleteFilesImpl (BYTE noItems)
{
  wchar_t tempDirName[300];
  if(GetTempPathW(299, tempDirName) == 0) lstrcpyW(tempDirName, L".");

  wchar_t tempFileName[300];
  if(GetTempFileNameW(tempDirName, L"scl", 0, tempFileName) == 0) return false;
  
  if(!openHostFile()) return false;
  
  HANDLE tempFile = CreateFileW(tempFileName,
                               GENERIC_WRITE,
                               FILE_SHARE_READ | FILE_SHARE_WRITE,
                               NULL,
                               CREATE_ALWAYS	,
                               FILE_FLAG_SEQUENTIAL_SCAN,
                               NULL);
  if(tempFile == INVALID_HANDLE_VALUE)
  {
    errorMessage(GetMsg(MCanNotCreateTmp));
    closeHostFile();
    return false;
  }
  
  DWORD checkSum       = writeSCLHeader(tempFile, noFiles - noItems);
  BYTE  noFilesWritten = 0;
  
  DWORD noBytesWritten;
  for(int i = 0; i < noFiles; ++i)
  {
    if(!pcFiles[i].deleted)
    {
      //  
      SetFilePointer(tempFile, noFilesWritten*sizeof(SCLHdr)+8+1, NULL, FILE_BEGIN);
      WriteFile(tempFile, &files[i], sizeof(SCLHdr), &noBytesWritten, NULL);
      checkSum += calculateCheckSum((BYTE*)&files[i], sizeof(SCLHdr));
      
      checkSum += copyFile(i, tempFile);
      ++noFilesWritten;
    }
  }
  //   
  WriteFile(tempFile, &checkSum, sizeof(checkSum), &noBytesWritten, NULL);
  
  CloseHandle(tempFile);
  closeHostFile();
  DeleteFileW(hostFileName);
  MoveFileW(tempFileName, hostFileName);
  op.reread = true;
  startupInfo.PanelControl(PANEL_PASSIVE, FCTL_UPDATEPANEL, 1, nullptr);
  op.reread = false;
  return true;
}

int Manager::deleteFiles(PluginPanelItem *panelItem, size_t noItems, int opMode)
{
  if(noItems == 0) return TRUE;
  if(isReadOnly)
  {
    const wchar_t *msgItems[4];
    msgItems [0] = GetMsg(MWarning);
    msgItems [1] = GetMsg(MCanNotDelete);
    msgItems [2] = GetMsg(MWriteProtected);
    msgItems [3] = GetMsg(MOk);
    messageBox(FMSG_WARNING, msgItems, std::size(msgItems), 1);
    return FALSE;
  }

  readInfo();

  if((opMode & OPM_SILENT) == 0)
  {
    const wchar_t *msgItems[5];
    msgItems[0] = GetMsg(MDelete);
    msgItems[1] = GetMsg(MDeleteAsk);
    msgItems[3] = GetMsg(MDeleteButton);
    msgItems[4] = GetMsg(MCancel);
    wchar_t msg[30];
    if(noItems == 1)
      wsprintfW(msg, L"%s", panelItem[0].FileName);
    else
      wsprintfW(msg, GetMsg(MDelFiles), noItems);
    
    msgItems[2] = msg;
    if(messageBox(0, msgItems, std::size(msgItems), 2)!=0)
      return FALSE;
    if(noItems > 1)
    {
      const wchar_t *msgItems[5];
      msgItems[0] = GetMsg(MDeleteFiles);
      msgItems[1] = GetMsg(MDeleteAsk);
      msgItems[3] = GetMsg(MAll);
      msgItems[4] = GetMsg(MCancel);

      wchar_t msg[30];
      wsprintfW(msg, GetMsg(MDelFiles), noItems);
      msgItems[2] = msg;
      if(messageBox(FMSG_WARNING, msgItems, std::size(msgItems), 2)!=0)
        return FALSE;
    }
  }
  
  //     
  int noFoundFiles = 0;
  for(int i = 0; i < noFiles; ++i)
  {
    for(int j = 0; j < noItems; ++j)
      if(!lstrcmpW(pcFiles[i].name, panelItem[j].FileName))
      {
        pcFiles[i].deleted = true;
        ++noFoundFiles;
        break;
      }
  }
  if(noFoundFiles != noItems) return FALSE;
  return (deleteFilesImpl(noItems) ? TRUE : FALSE);
}
