#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <wchar.h>

#include "plugin.hpp"

#include "types.hpp"
#include "types_scl.hpp"
#include "manager.hpp"
#include "detector.hpp"
#include "tools.hpp"

#include "debug.hpp"

extern Detector* detector;
extern Options op;

bool Manager::openHostFile(void)
{
  isReadOnly = false;
  DWORD mode = GENERIC_READ | GENERIC_WRITE;

  DWORD attr = GetFileAttributesW(hostFileName);
  if(attr & FILE_ATTRIBUTE_READONLY)
  {
    mode = GENERIC_READ;
    isReadOnly = true;
  }

  hostFile = CreateFileW(hostFileName,
                        mode,
                        FILE_SHARE_READ | FILE_SHARE_WRITE,
                        NULL,
                        OPEN_EXISTING,
                        FILE_FLAG_SEQUENTIAL_SCAN,
                        NULL);
  return (hostFile != INVALID_HANDLE_VALUE);
}

void Manager::closeHostFile(void)
{
  CloseHandle(hostFile);
}

void Manager::makePCNames(void)
{
  DWORD firstFileOffset = noFiles*sizeof(SCLHdr) + 8 + 1;
  for(int fNum = 0; fNum < noFiles; ++fNum)
  {
    const char* from    = files  [fNum].name;
    wchar_t* to      = pcFiles[fNum].name;

    makePCName(from, to);

    BYTE noChars = lstrlenW(to);

    BYTE dotPos = noChars;
    to[noChars++] = L'.';
    to[noChars++] = L'$';
    if(isValidChar(files[fNum].type)) to[noChars++] = Cp866ToUcs2(files[fNum].type);
    to[noChars] = 0;
    
    BYTE typeNum = 0xFF;
    if(op.detectFormat)
    {
      BYTE secs[16*256]{};

      int noSecs = files[fNum].noSecs;
      if(noSecs > 16) noSecs = 16;
      SetFilePointer(hostFile,
                   sectorSize*pcFiles[fNum].noSecsBefore+firstFileOffset,
                   NULL,
                   FILE_BEGIN);
    
      DWORD noBytesRead;
      ReadFile(hostFile, secs, noSecs*256, &noBytesRead, NULL);
      
      typeNum = detector->detect(files[fNum], secs, noSecs*256, pcFiles[fNum].comment);
    }
    pcFiles[fNum].type = typeNum;

    pcFiles[fNum].skipHeader = false;

    if(typeNum != 0xFF)
    {
      detector->specialChar(typeNum, to+dotPos+1);
      detector->getType(typeNum, to+dotPos+2);
      pcFiles[fNum].skipHeader = detector->getSkipHeader(typeNum);
    }
    //   zxzip 
    if(fNum != 0 &&
       !memcmp(from, "********ZIP", 11) &&
       !memcmp(&files[fNum-1].type, "ZIP", 3))
    {
      if(memcmp(files[fNum-1].name, "********", 8))
        lstrcpyW(to, pcFiles[fNum-1].name);
      else
        lstrcpynW(to, pcFiles[fNum-1].name, lstrlenW(pcFiles[fNum-1].name));
    }
    //   
    int i = fNum;
    while(i-- > 0)
    {
      int len = lstrlenW(to);
      if(!wcsnicmp(to, pcFiles[i].name, len))
      {
        BYTE ch = pcFiles[i].name[len];
        ch = (ch != 0) ? ch+1 : '0';
        
        //     :(
        if(ch == '9'+1) ch = 'A'; 
        if(ch == 'Z'+1) ch = 'a';
        if(ch == 'z'+1) ch = '0';
        
        to[len] = ch; to[len+1] = 0;
        break;
      }
    }
    pcFiles[fNum].deleted = false;
  }
}

bool Manager::readInfo(void)
{
  DWORD noBytesRead;
  
  //       
  WIN32_FIND_DATAW data;
  HANDLE h = FindFirstFileW(hostFileName, &data);
  if(h == INVALID_HANDLE_VALUE) return false;
  FindClose(h);
  
  if(CompareFileTime(&data.ftLastWriteTime, &lastModifed.ftLastWriteTime) != 0 ||
     data.nFileSizeLow != lastModifed.nFileSizeLow)
  {
    lastModifed = data;
    if(!openHostFile()) return false;

    SetFilePointer(hostFile, 8, NULL, FILE_BEGIN);
    ReadFile(hostFile, &noFiles, 1, &noBytesRead, NULL);
    if(noFiles == 0)
    {
      closeHostFile();
      return true;
    }

    DWORD totalSecs = 0;
    for(int i = 0; i < noFiles; ++i)
    {
      pcFiles[i].noSecsBefore = totalSecs;

      BYTE* p = (BYTE*)&files[i];
      ReadFile(hostFile, p, sizeof(SCLHdr), &noBytesRead, NULL);
      totalSecs += files[i].noSecs;
    }
    makePCNames();
    closeHostFile();
  }
  return true;
}

DWORD Manager::copyFile(int fileNum, HANDLE file, bool skipGarbage)
{
  if(!files[fileNum].noSecs) return 0;
  DWORD firstFileOffset = noFiles*sizeof(SCLHdr) + 8 + 1;
  SetFilePointer(hostFile,
                 sectorSize*pcFiles[fileNum].noSecsBefore+firstFileOffset,
                 NULL,
                 FILE_BEGIN);
  SetFilePointer(file, 0, NULL, FILE_END);
  
  DWORD noBytesWritten, noBytesRead;
  BYTE  sector[sectorSize];
  DWORD checkSum = 0;
  for(int s = 0; s < files[fileNum].noSecs-1; ++s)
  {
    ReadFile (hostFile, sector, sectorSize, &noBytesRead, NULL);
    WriteFile(file,     sector, sectorSize, &noBytesWritten, NULL);
    checkSum += calculateCheckSum(sector, sectorSize);
  }
  //     
  ReadFile(hostFile, sector, sectorSize, &noBytesRead, NULL);
  checkSum += calculateCheckSum(sector, sectorSize);
  int noBytesToWrite = sectorSize;
  int size           = files[fileNum].size;
  int noSecs2        = size/sectorSize + (size%sectorSize ? 1 : 0);
  if(skipGarbage && files[fileNum].noSecs == noSecs2 && size%sectorSize)
    noBytesToWrite = size % sectorSize;
  WriteFile(file, sector, noBytesToWrite, &noBytesWritten, NULL);
  return checkSum;
}
