#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdlib.h>
#include <memory>

#include "xTRD.hpp"
#include "types.hpp"
#include "types_trd.hpp"
#include "manager.hpp"
#include "fmtReader.hpp"
#include "detector.hpp"
#include "tools.hpp"
#include "lang.hpp"

PluginStartupInfo    startupInfo;
Options              op;

static FmtReader* fmtReader;
static HANDLE Settings;
Detector*  detector;

struct __declspec(uuid("{C2A8B4FB-4437-4C8B-80FB-FEF6AA207F06}")) TMainGuid {};
extern const GUID MainGuid = __uuidof(TMainGuid);

void WINAPI _export SetStartupInfoW(const struct PluginStartupInfo *info)
{
  if(info->StructSize < sizeof(PluginStartupInfo))
  {
      return;
  }
  //    PluginStartupInfo
  startupInfo = *info;

  FarSettingsCreate fsc { sizeof(FarSettingsCreate), MainGuid, 0 };
  startupInfo.SettingsControl(INVALID_HANDLE_VALUE, SCTL_CREATE, PSL_LOCAL, &fsc);
  Settings = fsc.Handle;

  FarSettingsItem fsi{ sizeof(FarSettingsItem), 0 };

  op.reread              = false;

  fsi.Name = L"ShowExt";
  fsi.Type = FST_QWORD;
  fsi.Number =  1;

  startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
  op.showExt             = fsi.Number;

  fsi.Name = L"DefaultPanelMode";
  fsi.Number =  5;
  startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
  op.defaultPanelMode    = fsi.Number;

  fsi.Name = L"DefaultFormat";
  fsi.Number =  1;
  startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
  op.defaultFormat       = fsi.Number;

  fsi.Name = L"DetectFormat";
  fsi.Number =  1;
  startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
  op.detectFormat        = fsi.Number;

  fsi.Name = L"AutoMove";
  fsi.Number =  1;
  startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
  op.autoMove            = fsi.Number;

  fsi.Name = L"UseDS";
  fsi.Number =  1;
  startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
  op.useDS               = fsi.Number;

  wchar_t cmdLine1[300] = L"c:\\dir1\\emul1.exe %P [%N]";
  wchar_t cmdLine2[300] = L"c:\\dir2\\emul2.exe %d \"%f\"";
  wchar_t types1[100] = L"Bb";
  wchar_t types2[100] = L"*";

  fsi.Type = FST_STRING;
  fsi.Name = L"CmdLine1";
  fsi.String = cmdLine1;
  startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
  wcsncpy(cmdLine1, fsi.String, std::size(cmdLine1));
  cmdLine1[std::size(cmdLine1) - 1] = 0;

  fsi.Name = L"CmdLine2";
  fsi.String = cmdLine2;
  startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
  wcsncpy(cmdLine2, fsi.String, std::size(cmdLine2));
  cmdLine2[std::size(cmdLine2) - 1] = 0;

  fsi.Name = L"Types1";
  fsi.String = types1;
  startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
  wcsncpy(types1, fsi.String, std::size(types1));
  types1[std::size(types1) - 1] = 0;

  fsi.Name = L"Types2";
  fsi.String = types2;
  startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
  wcsncpy(types2, fsi.String, std::size(types2));
  types2[std::size(types2) - 1] = 0;

  fsi.Name = L"FullScreen1";
  fsi.Type = FST_QWORD;
  fsi.Number =  1;
  startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
  op.fullScreen1 = fsi.Number;

  fsi.Name = L"FullScreen2";
  fsi.Number =  0;
  startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
  op.fullScreen2 = fsi.Number;

  fsi.Type = FST_STRING;
  fsi.Name = L"IniFilePath";
  fsi.String = L"";
  startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
  wcsncpy(op.iniFilePath, fsi.String, std::size(op.iniFilePath));
  op.iniFilePath[std::size(op.iniFilePath) - 1] = 0;


  detector            = new Detector (op.iniFilePath, startupInfo.ModuleName);
  fmtReader           = new FmtReader(startupInfo.ModuleName);
}

void WINAPI _export ExitFARW()
{
  delete detector;
  detector = nullptr;

  delete fmtReader;
  fmtReader = nullptr;

  startupInfo.SettingsControl(Settings, SCTL_FREE, 0, nullptr);
  Settings = nullptr;
}

HANDLE WINAPI _export AnalyseW(const struct AnalyseInfo *Info)
{
  if(Info->FileName == nullptr || Info->BufferSize == 0) return nullptr;
  
  const FmtPlugin* format = fmtReader->isImage(Info->FileName, (const BYTE *)Info->Buffer, Info->BufferSize);
  if(format == nullptr) return nullptr;
  
  Manager *m = new Manager(Info->FileName, format);
  return (HANDLE)m;
}

void WINAPI _export CloseAnalyseW(const struct CloseAnalyseInfo *Info)
{
    delete (Manager *)Info->Handle;
}

HANDLE WINAPI _export OpenW(const struct OpenInfo *Info)
{
  if(Info->OpenFrom == OPEN_ANALYSE)
  {
      OpenAnalyseInfo *An = (OpenAnalyseInfo *)Info->Data;
      return An->Handle;
  }

  if(Info->OpenFrom != OPEN_COMMANDLINE) return nullptr;
  const OpenCommandLineInfo *CommandLineInfo = (const OpenCommandLineInfo *)Info->Data;
  const wchar_t *fileName = CommandLineInfo->CommandLine;

  HANDLE hostFile = CreateFileW(fileName,
                               GENERIC_READ,
                               FILE_SHARE_READ | FILE_SHARE_WRITE,
                               NULL,
                               OPEN_EXISTING,
                               FILE_FLAG_SEQUENTIAL_SCAN,
                               NULL);
  if(hostFile == INVALID_HANDLE_VALUE) return nullptr;
  
  BYTE data[16*sectorSize];
  if(GetFileSize(hostFile, NULL) < sizeof(data))
  {
    CloseHandle(hostFile);
    return nullptr;
  }

  DWORD noBytesRead;
  ReadFile(hostFile, data, sizeof(data), &noBytesRead, NULL);
  CloseHandle(hostFile);
  
  const FmtPlugin* format = fmtReader->isImage(fileName, data, sizeof(data));
  if(format == NULL) return nullptr;
  
  Manager *m = new Manager(fileName, format);
  return (HANDLE)m;
}

void WINAPI _export ClosePanelW(const struct ClosePanelInfo *Info)
{
  delete (Manager*)Info->hPanel;
}

intptr_t  WINAPI _export GetFindDataW(struct GetFindDataInfo *Info)
{
  Manager* m = (Manager *)Info->hPanel;
  return m->getFindData(&Info->PanelItem, &Info->ItemsNumber, Info->OpMode);
}

void WINAPI _export FreeFindDataW(const struct FreeFindDataInfo *Info)
{
  Manager* m = (Manager*)Info->hPanel;
  m->freeFindData(Info->PanelItem, Info->ItemsNumber);
}

void WINAPI _export GetOpenPanelInfoW(struct OpenPanelInfo *Info)
{
  Manager* m = (Manager*)Info->hPanel;
  m->GetOpenPanelInfo(Info);
}

struct __declspec(uuid("{C82B957B-9FF9-4316-9E10-CBC81EBBA66B}")) TMenuGuid {};
static const GUID MenuGuid = __uuidof(TMenuGuid);

void WINAPI _export GetPluginInfoW(PluginInfo *info)
{
  info->StructSize = sizeof(*info);
  info->Flags      = PF_DISABLEPANELS;

  static const wchar_t *cmdPrefix = L"xTRD";
  info->CommandPrefix = cmdPrefix;
  static const wchar_t *pluginCfgStrings[1];
  pluginCfgStrings[0] = GetMsg(MConfig);

  info->PluginMenu.Guids = &MenuGuid;
  info->PluginMenu.Strings = pluginCfgStrings;
  info->PluginMenu.Count = std::size(pluginCfgStrings);

  info->PluginConfig.Guids = &MenuGuid;
  info->PluginConfig.Strings = pluginCfgStrings;
  info->PluginConfig.Count = std::size(pluginCfgStrings);
}

intptr_t  WINAPI _export ProcessPanelEventW(const struct ProcessPanelEventInfo *Info)
{
  if(Info->Event == FE_CHANGEVIEWMODE)
  {
    startupInfo.PanelControl(Info->hPanel, FCTL_UPDATEPANEL, 1, nullptr);
  }

  if(Info->Event == FE_CLOSE)
  {
    PanelInfo panel = { sizeof(PanelInfo) };
    startupInfo.PanelControl(Info->hPanel, FCTL_GETPANELINFO, 0, &panel);

    FarSettingsItem fsi{ sizeof(FarSettingsItem), 0 };

    op.defaultPanelMode = panel.ViewMode;
    fsi.Type = FST_QWORD;
    fsi.Name = L"DefaultPanelMode";
    fsi.Number =  panel.ViewMode;
    startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);
  }

  return FALSE;
}

intptr_t  WINAPI _export SetDirectoryW(const struct SetDirectoryInfo *Info)
{
  Manager* m = (Manager*)Info->hPanel;
  return m->setDirectory(Info->Dir, Info->OpMode);
}

intptr_t  WINAPI _export MakeDirectoryW(struct MakeDirectoryInfo *Info)
{
  Manager* m = (Manager*)Info->hPanel;
  return m->makeDirectory(Info->Name, Info->OpMode);
}

intptr_t WINAPI _export DeleteFilesW(const struct DeleteFilesInfo *Info)
{
  Manager* m = (Manager*)Info->hPanel;
  return (m->deleteFiles(Info->PanelItem, Info->ItemsNumber, Info->OpMode));
}

intptr_t WINAPI _export GetFilesW(struct GetFilesInfo *Info)
{
  Manager* m = (Manager*)Info->hPanel;
  return m->getFiles(Info->PanelItem, Info->ItemsNumber, Info->Move, Info->DestPath, Info->OpMode);
}

intptr_t WINAPI _export PutFilesW(const struct PutFilesInfo *Info)
{
  Manager* m = (Manager*)Info->hPanel;
  return (m->putFiles(Info->PanelItem, Info->ItemsNumber, Info->Move, Info->OpMode));
}

intptr_t WINAPI _export ConfigureW(const struct ConfigureInfo *Info)
{
  FarDialogItem dlgItems[] =
  {
    { DI_DOUBLEBOX,   3, 1, 60, 23, { FALSE },           nullptr, nullptr, 0,                                   GetMsg(MConfig),        0, 0, { 0, 0 } },
    { DI_CHECKBOX,    5, 3,  0,  0, { op.showExt },      nullptr, nullptr, 0,                                   GetMsg(MShowExt),       0, 0, { 0, 0 } },
    { DI_CHECKBOX,    5, 2,  0,  0, { op.detectFormat }, nullptr, nullptr, 0,                                   GetMsg(MDetectFormats), 0, 0, { 0, 0 } },
    { DI_CHECKBOX,    5, 4,  0,  0, { op.autoMove },     nullptr, nullptr, 0,                                   GetMsg(MAutoMove),      0, 0, { 0, 0 } },
    { DI_CHECKBOX,    5, 5,  0,  0, { op.useDS },        nullptr, nullptr, 0,                                   GetMsg(MUseDS),         0, 0, { 0, 0 } },
    { DI_TEXT,        3, 6,  0,  0, { FALSE },           nullptr, nullptr, DIF_BOXCOLOR | DIF_SEPARATOR,        L"",                    0, 0, { 0, 0 } },
    { DI_RADIOBUTTON, 5, 7,  0,  0, { FALSE },           nullptr, nullptr, DIF_GROUP,                           GetMsg(MHoBeta),        0, 0, { 0, 0 } },
    { DI_RADIOBUTTON, 5, 8,  0,  0, { FALSE },           nullptr, nullptr, 0,                                   GetMsg(MSCL),           0, 0, { 0, 0 } },
    { DI_RADIOBUTTON, 5, 9,  0,  0, { FALSE },           nullptr, nullptr, 0,                                   GetMsg(MAutoSelect),    0, 0, { 0, 0 } },
    { DI_TEXT,        3,10,  0,  0, { FALSE },           nullptr, nullptr, DIF_BOXCOLOR | DIF_SEPARATOR,        L"",                    0, 0, { 0, 0 } },
    { DI_TEXT,        5,11,  0,  0, { FALSE },           nullptr, nullptr, 0,                                   GetMsg(MEnterExec),     0, 0, { 0, 0 } },
    { DI_TEXT,        5,12,  0,  0, { FALSE },           nullptr, nullptr, 0,                                   GetMsg(MTypes),         0, 0, { 0, 0 } },
    { DI_EDIT,       19,12, 58,  0, { FALSE },           nullptr, nullptr, 0,                                   op.types1,              0, 0, { 0, 0 } },
    { DI_TEXT,        5,13,  0,  0, { FALSE },           nullptr, nullptr, 0,                                   GetMsg(MCmdLine),       0, 0, { 0, 0 } },
    { DI_EDIT,       19,13, 58,  0, { FALSE },           nullptr, nullptr, 0,                                   op.cmdLine1,            0, 0, { 0, 0 } },
    { DI_CHECKBOX,   19,14,  0,  0, { op.fullScreen1 },  nullptr, nullptr, 0,                                   GetMsg(MFullScr),       0, 0, { 0, 0 } },
    { DI_TEXT,        5,15,  0,  0, { FALSE },           nullptr, nullptr, 0,                                   GetMsg(MCtrlEnterExec), 0, 0, { 0, 0 } },
    { DI_TEXT,        5,16,  0,  0, { FALSE },           nullptr, nullptr, 0,                                   GetMsg(MTypes),         0, 0, { 0, 0 } },
    { DI_EDIT,       19,16, 58,  0, { FALSE },           nullptr, nullptr, 0,                                   op.types2,              0, 0, { 0, 0 } },
    { DI_TEXT,        5,17,  0,  0, { FALSE },           nullptr, nullptr, 0,                                   GetMsg(MCmdLine),       0, 0, { 0, 0 } },
    { DI_EDIT,       19,17, 58,  0, { FALSE },           nullptr, nullptr, 0,                                   op.cmdLine2,            0, 0, { 0, 0 } },
    { DI_CHECKBOX,   19,18,  0,  0, { op.fullScreen2 },  nullptr, nullptr, 0,                                   GetMsg(MFullScr),       0, 0, { 0, 0 } },
    { DI_TEXT,        3,19,  0,  0, { FALSE },           nullptr, nullptr, DIF_BOXCOLOR | DIF_SEPARATOR,        L"",                    0, 0, { 0, 0 } },
    { DI_TEXT,        5,20,  0,  0, { FALSE },           nullptr, nullptr, 0,                                   GetMsg(MIniPath),       0, 0, { 0, 0 } },
    { DI_EDIT,       25,20, 58,  0, { FALSE },           nullptr, nullptr, 0,                                   op.iniFilePath,         0, 0, { 0, 0 } },
    { DI_TEXT,        3,21,  0,  0, { FALSE },           nullptr, nullptr, DIF_BOXCOLOR | DIF_SEPARATOR,        L"",                    0, 0, { 0, 0 } },
    { DI_BUTTON,      0,22,  0,  0, { FALSE },           nullptr, nullptr, DIF_CENTERGROUP | DIF_DEFAULTBUTTON, GetMsg(MSave),          0, 0, { 0, 0 } },
    { DI_BUTTON,      0,22,  0,  0, { FALSE },           nullptr, nullptr, DIF_CENTERGROUP,                     GetMsg(MCancel),        0, 0, { 0, 0 } },
  };
  
  int num = 5 + op.defaultFormat;
  if(num < 6 || num > 8) num = 8;
  dlgItems[num].Selected = TRUE;

  std::unique_ptr<void, decltype(SafeDlgFree) *> Dlg{ startupInfo.DialogInit(&MainGuid,
                                   &__uuidof(struct __declspec(uuid("{0FFE73F2-75F3-43C2-8AD5-147A39DF24BE}")) TDlgGuid),
                                   -1, -1, 64, 25,
                                   nullptr,
                                   dlgItems,
                                   std::size(dlgItems), 0, 0, startupInfo.DefDlgProc, nullptr), SafeDlgFree };
  if(Dlg.get() == INVALID_HANDLE_VALUE)
  {
      return FALSE;
  }
  intptr_t askCode = startupInfo.DialogRun(Dlg.get());

  if(askCode != 26)
  {
      return FALSE;
  }

  op.detectFormat = startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 1, 0);
  op.showExt      = startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 2, 0);
  op.autoMove     = startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 3, 0);
  op.useDS        = startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 4, 0);
  
  op.defaultFormat = 3;
  if(startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 6, 0)) op.defaultFormat = 1;
  if(startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 7, 0)) op.defaultFormat = 2;

  FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(op.types1) - 1, op.types1 };
  startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 12, &ItemData);

  ItemData = { sizeof(FarDialogItemData), std::size(op.cmdLine1) - 1, op.cmdLine1 };
  startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 14, &ItemData);

  op.fullScreen1     = startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 15, 0);
  
  ItemData = { sizeof(FarDialogItemData), std::size(op.types2) - 1, op.types2 };
  startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 18, &ItemData);

  ItemData = { sizeof(FarDialogItemData), std::size(op.cmdLine2) - 1, op.cmdLine2 };
  startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 20, &ItemData);

  op.fullScreen2     = startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 21, 0);

  ItemData = { sizeof(FarDialogItemData), std::size(op.iniFilePath) - 1, op.iniFilePath };
  startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 24, &ItemData);

  FarSettingsItem fsi{ sizeof(FarSettingsItem), 0 };

  fsi.Type = FST_QWORD;

  fsi.Name = L"DetectFormat";
  fsi.Number =  op.detectFormat;
  startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);

  fsi.Name = L"ShowExt";
  fsi.Number =  op.showExt;
  startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);

  fsi.Name = L"DefaultFormat";
  fsi.Number =  op.defaultFormat;
  startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);

  fsi.Name = L"AutoMove";
  fsi.Number =  op.autoMove;
  startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);

  fsi.Name = L"UseDS";
  fsi.Number =  op.useDS;
  startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);

  fsi.Name = L"FullScreen1";
  fsi.Number =  op.fullScreen1;
  startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);

  fsi.Name = L"FullScreen2";
  fsi.Number =  op.fullScreen2;
  startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);

  fsi.Type = FST_STRING;

  fsi.Name = L"CmdLine1";
  fsi.String =  op.cmdLine1;
  startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);

  fsi.Name = L"CmdLine2";
  fsi.String =  op.cmdLine2;
  startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);

  fsi.Name = L"Types1";
  fsi.String =  op.types1;
  startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);

  fsi.Name = L"Types2";
  fsi.String =  op.types2;
  startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);

  fsi.Name = L"IniFilePath";
  fsi.String =  op.iniFilePath;
  startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);

  return TRUE;
}

intptr_t WINAPI _export ProcessPanelInputW(const struct ProcessPanelInputInfo *Info)
{
  Manager* m = (Manager*)Info->hPanel;
  return m->processKey(&Info->Rec);
}

intptr_t WINAPI _export ProcessHostFileW(const struct ProcessHostFileInfo *Info)
{
  Manager* m = (Manager*)Info->hPanel;
  return (m->processHostFile(Info->PanelItem, Info->ItemsNumber, Info->OpMode));
}

void WINAPI _export GetGlobalInfoW(struct GlobalInfo *Info)
{
    Info->StructSize = sizeof(struct GlobalInfo);
    Info->MinFarVersion = MAKEFARVERSION(3, 0, 0, 2927, VS_RELEASE);   //   Far Manager,    API
    Info->Version = MAKEFARVERSION(1, 0, 0, 1, VS_RC);   //   1.0.0.1, -
    Info->Guid = MainGuid;
    Info->Title = L"xTRD";
    Info->Description = L"xTRD";
    Info->Author = L"Elf/2 & Alexandr Medvedev";
}
